use crate::core::*;
use crate::func::*;
use crate::std::tracing::*;

use super::*;

pub struct TestContextTraced;

impl Context for TestContextTraced {
    type T = TracedClass;

    type D = TracedDiagnostic;

    type LookupError<'a> = TestLookupError<'a>;

    fn hash(s: &[u8]) -> Hash {
        TestContextPlain::hash(s)
    }
}

struct TracedResolver<'a> {
    resolver: Rc<dyn Resolver<'a, TestContextTraced>>,
}

impl<'a> TracedResolver<'a> {
    fn new(
        resolver: Rc<dyn Resolver<'a, TestContextTraced>>,
    ) -> Rc<dyn Resolver<'a, TestContextTraced>> {
        Rc::new(Self { resolver })
    }
}

impl<'a> Resolver<'a, TestContextTraced> for TracedResolver<'a> {
    fn resolve(self: Rc<Self>, address: Address) -> HashResolution<'a, TestContextTraced> {
        TracedClass::fmap(
            |resolved| {
                let (src, resolver) = resolved?;
                let delayed: Rc<dyn Resolver<'a, TestContextTraced>> =
                    Rc::new(TracedResolver { resolver });
                Ok((src, delayed))
            },
            self.resolver.clone().resolve(address),
        )
        .after_resolution()
    }
}

pub trait Traceable<'a>: Mentionable<'a, TestContextTraced> + Sized {
    fn trace(self: Rc<Self>) -> CastResult<'a, TestContextTraced, Self>;
}

impl<'a, A: Mentionable<'a, TestContextTraced>> Traceable<'a> for A {
    fn trace(self: Rc<Self>) -> CastResult<'a, TestContextTraced, Self> {
        let factory = self.factory();
        TypelessMentionable::from_typed(self).cast_full(factory, TracedResolver::new)
    }
}
